#!/bin/sh

PREFIX="[$0]"
FW_ARCHIVE=$1
NO_REBOOT="0"
CHECK_ACTUAL_VERSION="0"
FILE_LOCK_UPGRADE=/tmp/lock_multiple_upgrade
BOOT_VERSION_FILE=/tmp/boot_version_file
BANK_ACTIVE=
BANK_NEXT=
BANK_1_COMMIT=0
BANK_2_COMMIT=0
BANK_1_TRY=3
BANK_2_TRY=3
BOOTLOADER_IS_FOUND=0
FIRMWARE_IS_FOUND=0
BOOTLOADER_INSTALL_REQUIRED=0
FIRMWARE_INSTALL_REQUIRED=0

BOOTLOADER_MTD=0
ENV_MTD=2
ENV_BANK_OFFSET=0x40000
BANK_1_START_MTD=4
BANK_2_START_MTD=7
ENV_SIZE_IN_BLOCKS=2

FILE_BOOT="bootloader.bin"
FILE_ENV="u-boot-env.bin"
FILE_DTB="dtb.bin"
FILE_KERNEL="kernel.bin"
FILE_ROOTFS="rootfs.bin"

FIRMWARE_MTD_LIST="$FILE_DTB $FILE_KERNEL $FILE_ROOTFS"

ARCH_INFO=

create_lock_file() {
	echo "$PREFIX [INFO] Creating $FILE_LOCK_UPGRADE" > /dev/console
	echo "$0" > $FILE_LOCK_UPGRADE
}

# $1 - exit code
process_exit() {
  code=$1

  if [ "$code" -ne "0" ]; then
    rm ${FILE_LOCK_UPGRADE}
    echo "$PREFIX [INFO] $FILE_LOCK_UPGRADE was deleted" > /dev/console
  fi

  exit $code
}

read_env_variables() {
	echo "$PREFIX [INFO] Reading variables..." > /dev/console

	BANK_ACTIVE=$(($(ubootenv -g "bank_active" | cut -d '=' -f2)))
	BANK_1_TRY=$(($(ubootenv -g "bank1_try" | cut -d '=' -f2)))
	BANK_2_TRY=$(($(ubootenv -g "bank2_try" | cut -d '=' -f2)))
	BANK_1_COMMIT=$(($(ubootenv -g "bank1_commit" | cut -d '=' -f2)))
	BANK_2_COMMIT=$(($(ubootenv -g "bank2_commit" | cut -d '=' -f2)))

	echo "$PREFIX [INFO]   bank_active is $BANK_ACTIVE" > /dev/console
	echo "$PREFIX [INFO]   bank1_try is $BANK_1_TRY" > /dev/console
	echo "$PREFIX [INFO]   bank2_try is $BANK_2_TRY" > /dev/console
	echo "$PREFIX [INFO]   bank1_commit is $BANK_1_COMMIT" > /dev/console
	echo "$PREFIX [INFO]   bank2_commit is $BANK_2_COMMIT" > /dev/console
}

save_env_variables() {
	echo "$PREFIX [INFO] Saving variables..." > /dev/console

	ubootenv -s bank_active $BANK_ACTIVE
	ubootenv -s bank1_try $BANK_1_TRY
	ubootenv -s bank2_try $BANK_2_TRY
	ubootenv -s bank1_commit $BANK_1_COMMIT
	ubootenv -s bank2_commit $BANK_2_COMMIT
}

# $1 arg == mtd
# $2 arg == type (0 == write, 1 == erase)
log_failed() {
	MTD=$1
	TYPE=$2

	if [[ "$TYPE" == "0" ]]
	then
		echo "$PREFIX [INFO] Writing $MTD - FAILED" > /dev/console
	else
		echo "$PREFIX [INFO] Erasing $MTD - FAILED" > /dev/console
	fi

	echo "$PREFIX [ERROR] Firmware upgrade has been FAILED" > /dev/console
	process_exit 1
}

# $1 arg == mtd
# $2 type (0 == write, 1 == erase)
log_ok() {
	MTD=$1
	TYPE=$2

	if [[ "$TYPE" == "0" ]]
	then
		echo "$PREFIX [INFO] Writing $MTD - OK" > /dev/console
	else
		echo "$PREFIX [INFO] Erasing $MTD - OK" > /dev/console
	fi
}

# $1 arg == mtd number
erase_mtd() {
	MTD="/dev/mtd$1"

	MEM=$(($(cat /proc/mtd | grep "mtd$1" | awk '{pre = "0x"; print pre $2 }')))
	BLOCK_SIZE=$(($(cat /proc/mtd | grep "mtd$1" | awk '{pre = "0x"; print pre $3 }')))
	BLOCK_COUNT=$(expr $MEM / $BLOCK_SIZE)

	echo "$PREFIX [INFO] Erasing $MTD..." > /dev/console

	flash_erase $MTD 0 $BLOCK_COUNT > /dev/null

	if [ "$?" -ne "0" ]; then
		log_failed $MTD 1
	else
		log_ok $MTD 1
	fi
}

# $1 arg == mtd number
# $2 arg == path to file
write_mtd() {
	MTD="/dev/mtd$1"
	WRITE_FILE=$2
	IMG_SIZE=$( echo "$ARCH_INFO" | grep -w "$WRITE_FILE" | awk '{print $3}' )

	echo "$PREFIX [INFO] Writing $WRITE_FILE size $IMG_SIZE to $MTD..." > /dev/console

	tar -xf $FW_ARCHIVE $WRITE_FILE -O | nandwrite -a -k -m -p --input-size=$IMG_SIZE $MTD -

	if [ "$?" -ne "0" ]; then
		log_failed $MTD 0
	else
		log_ok $MTD 0
	fi
}

# $1 arg == mtd number
# $2 arg == BANK_ENV
erase_env() {
	MTD="/dev/mtd$1"
	BANK_ENV=$2

	if [ "$BANK_ENV" -eq "1" ]; then
		OFFSET=0
	else
		OFFSET=$ENV_BANK_OFFSET
	fi

	echo "$PREFIX [INFO] Erasing $MTD at 0x$OFFSET with $ENV_SIZE_IN_BLOCKS blocks..." > /dev/console

	flash_erase $MTD $OFFSET $ENV_SIZE_IN_BLOCKS > /dev/null

	if [ "$?" -ne "0" ]; then
		log_failed $MTD 1
	else
		log_ok $MTD 1
	fi
}

# $1 arg == mtd number
# $2 arg == BANK_ENV
write_env() {
	MTD="/dev/mtd$1"
	BANK_ENV=$2
	IMG_SIZE=$(tar -ztvf $FW_ARCHIVE $FILE_ENV | awk '{print $3}')

	if [ "$BANK_ENV" -eq "1" ]; then
		OFFSET=0
		BANK_NAME="env bank 1"
	else
		OFFSET=$ENV_BANK_OFFSET
		BANK_NAME="env bank 2"
	fi

	echo "$PREFIX [INFO] Writing $FILE_ENV size $IMG_SIZE to $BANK_NAME..." > /dev/console

	tar -xf $FW_ARCHIVE $FILE_ENV -O | nandwrite -a -k -m -N -s $OFFSET --input-size=$IMG_SIZE $MTD -
	if [ "$?" -ne "0" ]; then
		log_failed $MTD 0
	else
		log_ok $MTD 0
	fi
}

check_files_exist() {
	echo "$PREFIX [INFO] Check files exist" > /dev/console

	ARCH_INFO=$(tar -ztvf $FW_ARCHIVE)

	BOOT_EXIST=$( echo "$ARCH_INFO" | grep -wq "$FILE_BOOT" && echo 1 || echo 0 )
	ENV_EXIST=$( echo "$ARCH_INFO" | grep -wq "$FILE_ENV" && echo 1 || echo 0 )
	DTB_EXIST=$( echo "$ARCH_INFO" | grep -wq "$FILE_DTB" && echo 1 || echo 0 )
	KERNEL_EXIST=$( echo "$ARCH_INFO" | grep -wq "$FILE_KERNEL" && echo 1 || echo 0 )
	ROOTFS_EXIST=$( echo "$ARCH_INFO" | grep -wq "$FILE_ROOTFS" && echo 1 || echo 0 )

	if [ "$BOOT_EXIST" -eq "1" ] && [ "$ENV_EXIST" -eq "1" ]; then
		BOOTLOADER_IS_FOUND=1
		BOOTLOADER_INSTALL_REQUIRED=1
	fi

	if [ "$DTB_EXIST" -eq "1" ] && [ "$KERNEL_EXIST" -eq "1" ] && [ "$ROOTFS_EXIST" -eq "1" ]; then
		FIRMWARE_IS_FOUND=1
		FIRMWARE_INSTALL_REQUIRED=1
	fi

	if [ ! -f "./version" ]; then
		echo "$PREFIX [ERROR] ./version doesn't exist!" > /dev/console
		process_exit 1
	fi

	if [ "$FIRMWARE_IS_FOUND" -eq "0" ] && [ "$BOOTLOADER_IS_FOUND" -eq "0" ]; then
		echo "$PREFIX [ERROR] Archive content isn't correct!" > /dev/console
		process_exit 1
	fi
}

check_actual_version() {
	if [ "$FIRMWARE_IS_FOUND" -eq "1" ] ; then
		cmp -s version /etc/version
	
		if [ "$?" == 0 ]; then
			echo "$PREFIX [WARNING] The actual firmware version is already installed" > /dev/console
			FIRMWARE_INSTALL_REQUIRED=0
		fi
	fi
	
	if [ "$BOOTLOADER_IS_FOUND" -eq "1" ] ; then
		ubootenv -g ver | cut -d '=' -f 2 > $BOOT_VERSION_FILE
		cmp -s version $BOOT_VERSION_FILE

		if [ "$?" == 0 ]; then
			echo "$PREFIX [WARNING] The actual bootloader version is already installed" > /dev/console
			BOOTLOADER_INSTALL_REQUIRED=0
		fi

		rm $BOOT_VERSION_FILE
	fi
}

check_version() {
	if [[ $CHECK_ACTUAL_VERSION -ne 0 ]]; then
		echo "$PREFIX [INFO] Check actual firmware version" > /dev/console
		check_actual_version
	fi
}

check_upgrade_needing() {
	echo "$PREFIX [INFO] Check upgrade needing" > /dev/console
	check_version

	if [ "$FIRMWARE_IS_FOUND" -eq "1" ] && [ "$BOOTLOADER_IS_FOUND" -eq "0" ]; then
		if [ "$FIRMWARE_INSTALL_REQUIRED" -eq "0" ]; then
			process_exit 2
		fi
	fi

	if [ "$FIRMWARE_IS_FOUND" -eq "0" ] && [ "$BOOTLOADER_IS_FOUND" -eq "1" ]; then
		if [ "$BOOTLOADER_INSTALL_REQUIRED" -eq "0" ]; then
			process_exit 2
		fi
	fi

	if [ "$FIRMWARE_IS_FOUND" -eq "1" ] && [ "$BOOTLOADER_IS_FOUND" -eq "1" ]; then
		if [ "$FIRMWARE_INSTALL_REQUIRED" -eq "0" ] && [ "$BOOTLOADER_INSTALL_REQUIRED" -eq "0" ]; then
			process_exit 2
		fi
	fi

	VERSION="$(cat ./version)"
	echo "$PREFIX [INFO] Version $VERSION will be installed" > /dev/console
}

upgrade_firmware() {
	ITER=0

	if [ "$BANK_NEXT" -eq "1" ]; then
		START_MTD=$BANK_1_START_MTD
	else
		START_MTD=$BANK_2_START_MTD
	fi

	for IMAGE in $FIRMWARE_MTD_LIST; do
		MTD_NUM=$(($START_MTD + $ITER))

		erase_mtd ${MTD_NUM}
		write_mtd ${MTD_NUM} $IMAGE

		ITER=$((ITER+1))
	done
}

upgrade_bootloader() {
	erase_mtd $BOOTLOADER_MTD
	write_mtd $BOOTLOADER_MTD $FILE_BOOT

	erase_env $ENV_MTD 1
	write_env $ENV_MTD 1

	erase_env $ENV_MTD 2
	write_env $ENV_MTD 2
}

upgrade_process() {
	#Step 1. Checking
	check_files_exist
	check_upgrade_needing

	#Step 2. Send upgrade notification
	type upgrade-notification >/dev/null

	if [ "$?" -eq "0" ]; then
		upgrade-notification
		NOTIFY_STATUS=$?

		if [ "$NOTIFY_STATUS" -eq "5" ]; then
			echo "$PREFIX [ERROR] There are active calls on phone, upgrade aborted" > /dev/console
			process_exit 5
		fi
	fi

	systemctl stop voip-app

	#Step 3. Start the upgrade process
	if [ "$BOOTLOADER_INSTALL_REQUIRED" -eq "1" ]; then
		echo "===== Start of BOOTLOADER upgrade =====" > /dev/console

		read_env_variables

		upgrade_bootloader

		echo "$PREFIX [INFO] Recovering variables" > /dev/console

		save_env_variables

		echo "===== End of BOOTLOADER upgrade =====" > /dev/console
	fi

	if [ "$FIRMWARE_INSTALL_REQUIRED" -eq "1" ]; then
		echo "===== Start of FIRMWARE upgrade =====" > /dev/console

		read_env_variables

		if [ "$BANK_ACTIVE" -eq "1" ]; then
			BANK_2_COMMIT=0
			BANK_NEXT=2
		else
			BANK_1_COMMIT=0
			BANK_NEXT=1
		fi
		save_env_variables

		upgrade_firmware

		if [ "$BANK_ACTIVE" -eq "1" ]; then
			BANK_2_TRY=3
			BANK_ACTIVE=2
		else
			BANK_1_TRY=3
			BANK_ACTIVE=1
		fi
		save_env_variables

		echo "===== End of FIRMWARE upgrade =====" > /dev/console
	fi

	echo "$PREFIX [INFO] Upgrade has been successfully completed. Well done!" > /dev/console
}

###############################################################################################

if [ -f $FILE_LOCK_UPGRADE ]; then
	echo "$PREFIX [ERROR] $FILE_LOCK_UPGRADE has been detected!" > /dev/console
	exit 3
fi

shift

while getopts nv flag
do
	case "${flag}" in
		n) NO_REBOOT="1";;
		v) CHECK_ACTUAL_VERSION="1";;
	esac
done

create_lock_file

upgrade_process

if [[ $NO_REBOOT -ne 1 ]]; then
	echo "$PREFIX [INFO] The system will be rebooted" > /dev/console
	reboot
fi
